/*
   Watch5 ARM9 Code: fragments.c
   Maarten Pennings - 2007 dec 25
   Split-off from arm9.c: the initialization of the TransferSoundData struct's
   for 20ish sound fragments.
*/

/* This file uses a trick: fragments.i is included 3 times.
** File fragments.i contains a list of all sound fragments (their name and their meta data).
** All data (name, meta data) of a FRAGMENT are captured on a single line and tagged with FRAGMENT.
** The includes are used to generate
**  1) the symbols to the sound data (pointer to start of PCM binary)
**  2) an enum identifying all sounds
**  3) initialization code for an array describing all sounds
** In order to achieve different behavior (for the three #includes), the macro FRAGMENT
** must be set, before each #include.
*/

#include "nds.h"


// include #1 (making the xxx_bin.h superfluous)
#define FRAGMENT(id,rate_,vol_,pan_,fmt) extern const u8 id##_bin_end[];\
  extern const u8 id##_bin[];\
  extern const u32 id##_bin_size;
#include "fragments.i"
#undef FRAGMENT


// include 2 (to get the enum) which includes fragments.i
#include "fragments.h"


TransferSoundData aFragments[FRAGMENT_last];


// include #3
#define FRAGMENT(id,rate_,vol_,pan_,fmt) aFragments[FRAGMENT_##id].data=id##_bin;\
  aFragments[FRAGMENT_##id].len=id##_bin_size;\
  aFragments[FRAGMENT_##id].rate=rate_;\
  aFragments[FRAGMENT_##id].vol=vol_;\
  aFragments[FRAGMENT_##id].pan=pan_;\
  aFragments[FRAGMENT_##id].format=fmt;
void InitFragments( void ) {
#include "fragments.i"
}
#undef FRAGMENT
